#ifndef LINT
static char *rcsid="$Id: smscanner.c 297 2005-05-13 17:44:56Z crosser $";
#endif
                                                                                
/*
	WHAT IS IT:
		modularized contentfilter for Zmailer and Sendmail
	COPYRIGHT:
		(c) 2003-2005 Eugene G. Crosser <crosser@average.org>
	LICENSE:
		The same set as apply to Zmailer code
*/

/*
	developed using documentation from here:
		http://www.milter.org/milter_api/api.html
*/

#include "config.h"

#include <sys/types.h>
#ifdef STDC_HEADERS
# include <stdio.h>
# include <stdlib.h>
# include <stddef.h>
# include <string.h>
#endif
#ifdef HAVE_SYS_STAT_H
# include <sys/stat.h>
#endif
#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif
#ifdef HAVE_ERRNO_H
# include <errno.h>
#endif
#ifdef HAVE_FCNTL_H
# include <fcntl.h>
#endif
#include <signal.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/un.h>
#ifndef OPTARG_DEFINED
#include <getopt.h>
#endif

#include <libmilter/mfapi.h>

#include "rtconfig.h"
#include "report.h"
#include "zmscanner.h"
#include "daemon.h"

/* should be in some header file but I cannot find an appropriate one */
void init_plugin_hash(void);
void makepidfile(char *fn);

int verbose=0;
int isdaemon=1;

/* this frontend is only built in threaded version, no saveqfile anyway */
void saveqfile(void) {};

/* the thing does not work in milter:
RETSIGTYPE
terminate(int sig)
{
	DPRINT(("signal caught: %d\n",sig));
	(void)smfi_stop();
}
*/

/* Header scanners presume that text of the headers stay available for
   the whole time of message processing.  Content-Type, boundary, e.t.c.
   in the mattr structure point into the middle of the headers.
   Because milter does not give us the whole message, we'll have to keep
   all headers in a temporary chain.  It could be possible to only
   keep the chunks referenced by mattr, but that would require
   reimplementation of rfc822hdr scanner, and I don't want to touch
   code that is proved to work. */

typedef struct _hdr {
	struct _hdr *next;
	char text[4]; /* for ':', ' ', '\n' and trailing zero */
} hdr_t;

struct mlfiPriv
{
	varpool_t vp;
	hdr_t *firsthdr;
	hdr_t *lasthdr;
	mattr_t matr;
	int freeze;
	int notfirstbodychunk;
	int local;
	int msgcount;
	char *peerstr;
	char *helostr;
};

#define MLFIPRIV (mlfi_priv(ctx))

static struct mlfiPriv *
mlfi_priv(SMFICTX *ctx)
{
	struct mlfiPriv *priv = smfi_getpriv(ctx);

	if (priv) {
		DPRINT(("mlfi_priv: existing priv=%p\n",priv));
		return priv;
	}

	if ((priv=malloc(sizeof *priv)) == NULL) {
		ERRLOG((LOG_ERR,"mlfi_conn: malloc error: %m"));
		return NULL;
	}
	if ((priv->vp=vp_create_vp()) == NULL) {
		ERRLOG((LOG_ERR,"vp_create_vp failed: %m"));
		return NULL;
	}
	DPRINT(("mlfi_conn: varpool at %p\n",priv->vp));
	vp_set_str(priv->vp,VP_ANSWER_STR,"0 250 Acceptable");
	priv->firsthdr=NULL;
	priv->lasthdr=(hdr_t*)&(priv->firsthdr);
	memset(&(priv->matr),0,sizeof(priv->matr));
	/* in case that there is no content-type/content-encoding in
	   the message, fill in attribute block with default values */
	priv->matr.btype=t_single;
	priv->matr.benc=e_plain;
	priv->matr.usefirst=gcfg.firstmimehdr;
	priv->freeze=0;
	priv->notfirstbodychunk=0;
	priv->local=1;
	priv->msgcount=0;
	priv->peerstr=NULL;
	priv->helostr=NULL;
	smfi_setpriv(ctx, priv);
	DPRINT(("mlfi_priv: created priv=%p\n",priv));
	return priv;
}

sfsistat
mlfi_cleanup(SMFICTX *ctx,int ok)
{
	struct mlfiPriv *priv = MLFIPRIV;
	hdr_t *hdr;

	DPRINT(("mlfi_cleanup: priv=%p\n",priv));
	if (priv == NULL) return ok;

	for (hdr=priv->firsthdr;hdr;) {
		hdr_t *prev=hdr;
		hdr=hdr->next;
		free(prev);
	}
	vp_destroy_vp(priv->vp);
	if ((priv->vp=vp_create_vp()) == NULL) {
		ERRLOG((LOG_ERR,"vp_create_vp failed: %m"));
		return SMFIS_CONTINUE;
	}
	DPRINT(("mlfi_conn: varpool at %p\n",priv->vp));
	vp_set_str(priv->vp,VP_ANSWER_STR,"0 250 Acceptable");
	priv->firsthdr=NULL;
	priv->lasthdr=(hdr_t*)&(priv->firsthdr);
	memset(&(priv->matr),0,sizeof(priv->matr));
	/* in case that there is no content-type/content-encoding in
	   the message, fill in attribute block with default values */
	priv->matr.btype=t_single;
	priv->matr.benc=e_plain;
	priv->matr.usefirst=gcfg.firstmimehdr;
	priv->freeze=0;
	priv->notfirstbodychunk=0;
	priv->local=1;
	return ok;
}

static int
mlfi_termmode(SMFICTX *ctx,int stopcode)
{
	struct mlfiPriv *priv;
	char *ans;
	int rc;
	char rcode[4];
	char xcode[6];
	char *p,*q,*msg;

	if (stopcode == ZMSCAN_CONTINUE) return SMFIS_CONTINUE;
	priv=smfi_getpriv(ctx);
	if (priv == NULL) return SMFIS_CONTINUE;
	ans=vp_get_str(priv->vp,VP_ANSWER_STR);

	rc=atoi(ans);
	p=strchr(ans,' ');
	if (p) while (*p == ' ') p++;
	if (p && (strspn(p,"0123456789") == 3)) strncpy(rcode,p,3);
	else strncpy(rcode,"500",3);
	rcode[3]='\0';
	if (p) q=strchr(p,' ');
	else q=NULL;
	if (q) while (*q == ' ') q++;
	if (q && (strspn(q,"0123456789.") == 5)) strncpy(xcode,q,5);
	else xcode[0]='\0';
	xcode[5]='\0';
	if (xcode[0]) msg=q+5;
	else if (p) msg=p+3;
	else msg=NULL;
	if (msg) while (*msg == ' ') msg++;
	DPRINT(("mlfi_termmode: rc=%d rcode=\"%s\" xcode=\"%s\"\n",
			rc,rcode,xcode?xcode:"(nil)"));
	(void)smfi_setreply(ctx,rcode,xcode[0]?xcode:NULL,msg);

	switch (rc) {
	case 0:
		return SMFIS_ACCEPT;
	case 1:
#ifdef SMFIF_QUARANTINE
		priv->freeze=1;
		return SMFIS_CONTINUE;
#endif /* SMFIF_QUARANTINE; otherwise act as if it was REJECT */
	case -1:
		switch (rcode[0]) {
		case '5':	return SMFIS_REJECT;
		case '4':	return SMFIS_TEMPFAIL;
		default:	ERRLOG((LOG_ERR,"stage returned \"%s\" which "
				"contains unknown rcode %s",ans,rcode));
				return SMFIS_ACCEPT;	/* or what? */
		}
	default: ERRLOG((LOG_ERR,"stage returned \"%s\" which "
				"contains unknown rc %d",ans,rc));
		return SMFIS_ACCEPT;	/* something bad */
	}
}

sfsistat
mlfi_conn(SMFICTX *ctx,char *hostname,_SOCK_ADDR *hostaddr)
{
	slab_t data;
	int rc;
	char buf[128];
	char caddr[64];
	int port;
	struct mlfiPriv *priv = MLFIPRIV;

	DPRINT(("mlfi_conn: %s\n",hostname));
	if (priv == NULL) return SMFIS_TEMPFAIL;

	priv->local=0;
	switch (hostaddr->sa_family) {
	case AF_INET:
		inet_ntop(AF_INET,
			(void*)&((struct sockaddr_in*)hostaddr)->sin_addr,
			caddr,sizeof(caddr));
		port=((struct sockaddr_in*)hostaddr)->sin_port;
		break;
#ifdef AF_INET6
	case AF_INET6:
		inet_ntop(AF_INET6,
			(void*)&((struct sockaddr_in6*)hostaddr)->sin6_addr,
			caddr,sizeof(caddr));
		port=((struct sockaddr_in6*)hostaddr)->sin6_port;
		break;
#endif /* AF_INET6 */
	default:
		snprintf(caddr,sizeof(caddr),"-1 unknown family %d",
			hostaddr->sa_family);
		port=0;
	}
	snprintf(buf,sizeof(buf),"%s [%s]:%d",hostname,caddr,port);
	data.beg=buf;
	data.end=buf+strlen(buf);
	data.atr=(void *)&(priv->matr);
	DPRINT(("peerdata is \"%s\"\n",buf));
	if (priv->peerstr == NULL) {
		priv->peerstr=(char*)malloc(strlen(buf)+1);
	}
	if (priv->peerstr) strcpy(priv->peerstr,buf);
	rc=scandata("peerdata",0,data,priv->vp);
	return mlfi_termmode(ctx,rc);
}

sfsistat
mlfi_helo(SMFICTX *ctx,char *helohost)
{
	slab_t data;
	int rc;
	struct mlfiPriv *priv = MLFIPRIV;

	DPRINT(("mlfi_helo: %s\n",helohost));
	if (priv == NULL) return SMFIS_TEMPFAIL;
	if (priv->local) return SMFIS_CONTINUE;
	if (priv->freeze) return SMFIS_CONTINUE;
	data.beg=helohost;
	data.end=helohost+strlen(helohost);
	data.atr=(void *)&(priv->matr);
	if (priv->helostr == NULL) {
		priv->helostr=(char*)malloc(strlen(helohost)+1);
	}
	if (priv->helostr) strcpy(priv->helostr,helohost);
	rc=scandata("helo",0,data,priv->vp);
	return mlfi_termmode(ctx,rc);
}

sfsistat
mlfi_envfrom(SMFICTX *ctx,char **argv)
{
	slab_t data;
	int rc;
	struct mlfiPriv *priv = MLFIPRIV;

	DPRINT(("mlfi_envfrom: %s\n",argv[0]));
	if (priv == NULL) return SMFIS_TEMPFAIL;
	(priv->msgcount)++;
	if (priv->local) return SMFIS_CONTINUE;
	if (priv->freeze) return SMFIS_CONTINUE;
	if (priv->msgcount > 1) {
		if (priv->peerstr) {
			data.beg=priv->peerstr;
			data.end=(priv->peerstr)+strlen(priv->peerstr);
			data.atr=(void *)&(priv->matr);
			rc=scandata("peerdata",0,data,priv->vp);
		}
		if (priv->helostr) {
			data.beg=priv->helostr;
			data.end=(priv->helostr)+strlen(priv->helostr);
			data.atr=(void *)&(priv->matr);
			rc=scandata("helo",0,data,priv->vp);
		}
	}
	data.beg=argv[0];
	data.end=argv[0]+strlen(argv[0]);
	data.atr=(void *)&(priv->matr);
	rc=scandata("envfrom",0,data,priv->vp);
	return mlfi_termmode(ctx,rc);
}

sfsistat
mlfi_envrcpt(SMFICTX *ctx,char **argv)
{
	slab_t data;
	int rc;
	struct mlfiPriv *priv = MLFIPRIV;

	DPRINT(("mlfi_envrcpt: %s\n",argv[0]));
	if (priv == NULL) return SMFIS_TEMPFAIL;
	if (priv->local) return SMFIS_CONTINUE;
	if (priv->freeze) return SMFIS_CONTINUE;
	data.beg=argv[0];
	data.end=argv[0]+strlen(argv[0]);
	data.atr=(void *)&(priv->matr);
	rc=scandata("envrcpt",0,data,priv->vp);
	return mlfi_termmode(ctx,rc);
}

sfsistat
mlfi_header(SMFICTX *ctx,char *headerf,char *headerv)
{
	slab_t data;
	int rc;
	struct mlfiPriv *priv = MLFIPRIV;
	hdr_t *hdr;

	DPRINT(("mlfi_header: %s: %s\n",headerf,headerv));
	if (priv == NULL) return SMFIS_TEMPFAIL;
	if (priv->local) return SMFIS_CONTINUE;
	if (priv->freeze) return SMFIS_CONTINUE;
	if ((hdr=malloc(sizeof *hdr+strlen(headerf)+strlen(headerv))) == NULL) {
		ERRLOG((LOG_ERR,"mlfi_conn: malloc error: %m"));
		return mlfi_cleanup(ctx,SMFIS_TEMPFAIL);
	}
	hdr->next=NULL;
	sprintf(hdr->text,"%s: %s\n",headerf,headerv);
	priv->lasthdr->next=hdr;
	priv->lasthdr=hdr;
	data.beg=hdr->text;
	data.end=hdr->text+strlen(hdr->text);
	data.atr=(void *)&(priv->matr);
	DPRINT(("header is \"%s\"\n",hdr->text));
	rc=scandata("rfc822hdr",0,data,priv->vp);
	return mlfi_termmode(ctx,rc);
}

sfsistat
mlfi_eoh(SMFICTX *ctx)
{
	struct mlfiPriv *priv = MLFIPRIV;
	mattr_t *matr=&(priv->matr);

	DPRINT(("mlfi_eoh:\n"));
	/* matr should have been filled as a side effect of header scanning */
	DPRINT(("atr btype=%d\n",matr->btype));
	DPRINT(("atr benc=%d\n",matr->benc));
	DPRINT(("atr content_type=\"%.*s\"\n",
			slab_size(matr->content_type),matr->content_type.beg));
	DPRINT(("atr boundary=\"%.*s\"\n",
			slab_size(matr->boundary),matr->boundary.beg));
	DPRINT(("atr name=\"%.*s\"\n",
			slab_size(matr->name),matr->name.beg));
	DPRINT(("atr charset=\"%.*s\"\n",
			slab_size(matr->charset), matr->charset.beg));
	if ((matr->btype == t_multipart) && (slab_size(matr->boundary) == 0)) {
		/* I've seen DSNs from some broken MTA that generated
		   mulipart/mixed entities without boundary! Being
		   obvously broken, we have a choice of treating such
		   things as either text/plain or message/rfc822.
		   In my case, body of the entity was in fact an
		   enclosed message.  So, let's try this: */
		matr->btype=t_message;
	}

	return SMFIS_CONTINUE;
}

sfsistat
mlfi_body(SMFICTX *ctx,u_char *bodyp,size_t bodylen)
{
	slab_t data;
	int rc;
	struct mlfiPriv *priv = MLFIPRIV;

	DPRINT(("mlfi_body: size %d, %sfirst entry\n",bodylen,
			priv->notfirstbodychunk?"not ":""));
	if (priv == NULL) return SMFIS_TEMPFAIL;
	if (priv->local) return SMFIS_CONTINUE;
	if (priv->freeze) return SMFIS_CONTINUE;
	if (priv->notfirstbodychunk) return SMFIS_CONTINUE;
	priv->notfirstbodychunk=1;
	data.beg=bodyp;
	data.end=bodyp+bodylen;
	data.atr=(void *)&(priv->matr);
	rc=scandata("body",0,data,priv->vp);
	return mlfi_termmode(ctx,rc);
}

sfsistat
mlfi_eom(SMFICTX *ctx)
{
	struct mlfiPriv *priv = MLFIPRIV;
	sfsistat rc=SMFIS_CONTINUE;
	static char *nul="nul";
	slab_t data;

	DPRINT(("mlfi_eom\n"));
	if (priv == NULL) return SMFIS_TEMPFAIL;

	data.beg=nul;
	data.end=nul+strlen(nul);
	data.atr=(void *)&(priv->matr);
	(void)scandata(gcfg.finalstage,0,data,priv->vp);

#ifdef SMFIF_QUARANTINE
	if (priv->freeze) {
		struct mlfiPriv *priv = MLFIPRIV;
		char *ans=vp_get_str(priv->vp,VP_ANSWER_STR);
		char *p=strchr(ans,' ');

		if (p) while (*p == ' ') p++;
		else p="Frozen";
		(void)smfi_quarantine(ctx,p);
		rc=SMFIS_ACCEPT; /* is this right?  or what? */
	} else rc=mlfi_termmode(ctx,ZMSCAN_STOP);
#endif /* SMFIF_QUARANTINE */

	return mlfi_cleanup(ctx,rc);
}

sfsistat
mlfi_abort(SMFICTX *ctx)
{
	DPRINT(("mlfi_abort\n"));
	/* return code irrelevant here,message already aborted */
	return mlfi_cleanup(ctx,SMFIS_CONTINUE);
}

sfsistat
mlfi_close(SMFICTX *ctx)
{
	struct mlfiPriv *priv = smfi_getpriv(ctx);

	DPRINT(("mlfi_close\n"));
	(void)mlfi_cleanup(ctx,SMFIS_CONTINUE);
	if (priv->peerstr) free(priv->peerstr);
	if (priv->helostr) free(priv->helostr);
	free(priv);
	smfi_setpriv(ctx, NULL);
	return SMFIS_CONTINUE;
}

struct smfiDesc smfilter = {
	"smscanner",	/* filter name */
	SMFI_VERSION,	/* version code -- do not change */
	SMFIF_NONE
#ifdef SMFIF_QUARANTINE
	| SMFIF_QUARANTINE
#endif /* SMFIF_QUARANTINE */
		,	/* flags */
	mlfi_conn,	/* connection info filter */
	mlfi_helo,	/* SMTP HELO command filter */
	mlfi_envfrom,	/* envelope sender filter */
	mlfi_envrcpt,	/* envelope recipient filter */
	mlfi_header,	/* header filter */
	mlfi_eoh,	/* end of header */
	mlfi_body,	/* body block filter */
	mlfi_eom,	/* end of message */
	mlfi_abort,	/* message aborted */
	mlfi_close	/* connection cleanup */
};

static void
showver(char *name)
{
	fprintf(stderr,"%s Version %s Build %s\n",name,VERSION,__DATE__);
}

static void
showhelp(char *name)
{
	fprintf(stderr,"Usage: %s -I configfile -d -V -v -h\n",name);
	fprintf(stderr,"\t-I configfile\t- replacement config [default: %s]\n",DEFAULT_CONFIG);
	fprintf(stderr,"\t-p miltersock\t- replacement milter socket desc string\n");
	fprintf(stderr,"\t-d\t\t- do not go to background\n");
	fprintf(stderr,"\t-V\t\t- print version\n");
	fprintf(stderr,"\t-v\t\t- increase verbosity\n");
	fprintf(stderr,"\t-h\t\t- print this help message\n");
}

int
main(int argc, char *argv[])
{
	int i;
	int rc;
	char *configfile=DEFAULT_CONFIG;
	char *miltersock=NULL;

	while ((i=getopt(argc,argv,"I:p:dvVh")) != EOF) switch (i) {
	case 'I':
		/* if someone decides to modify the name displayed in
		   the "ps" output it may be useful to copy the arg... */
		configfile=(char *)malloc(strlen(optarg)+1);
		strcpy(configfile,optarg);
		break;
	case 'p':
		miltersock=(char *)malloc(strlen(optarg)+1);
		strcpy(miltersock,optarg);
		break;
	case 'd':
		isdaemon=0;
		break;
	case 'V':
		showver(argv[0]);
		return 0;
	case 'v':
		verbose++;
		break;
	case 'h':
		showhelp(argv[0]);
		return 0;
	default:
		showhelp(argv[0]);
		return 1;
	}

#ifdef HAVE_SYSLOG
	openlog("smscanner",LOG_CONS|LOG_PID
# ifdef LOG_NDELAY
					|LOG_NDELAY
# endif
						,LOG_MAIL);
#endif

	DPRINT(("process %d starting\n",getpid()));
	ERRLOG((LOG_INFO,"starting v.%s build %s %s",
					VERSION,__DATE__,__TIME__));
/* the thing does not work in milter:
	(void)signal(SIGTERM,terminate);
	(void)signal(SIGHUP,terminate);
	(void)signal(SIGINT,terminate);
*/
	init_plugin_hash();
	(void)readconfig(configfile,1);
	if (miltersock == NULL) miltersock=gcfg.miltersock;
	if (miltersock == NULL) {
		fprintf(stderr,"miltersock not specified\n");
		return 1;
	}

	if (isdaemon) {
		pid_t child;
		if ((child=fork()) > 0) {
			sleep(1);
			return 0;
		} else if (child < 0) {
			perror("fork error");
			return 1;
		} else {
			fflush(stdin);
			fflush(stdout);
			fflush(stderr);
			setbuf(stdin,NULL);
			setbuf(stdout,NULL);
			setbuf(stderr,NULL);
			close(0);
			close(1);
			close(2);
			open("/dev/null",O_RDWR);
			dup(0);
			dup(0);
#ifdef HAVE_SETSID
			setsid();
#else
# ifdef HAVE_SETPGRP
#  ifdef SETPGRP_VOID
			setpgrp();
#  else
			setpgrp(0,0);
#  endif
# endif
#endif
		}
	}

	if (smfi_setconn(miltersock) == MI_FAILURE) {
		ERRLOG((LOG_ERR,"smfi_setconn failed"));
		exit(1);
	}
	if (smfi_register(smfilter) == MI_FAILURE) {
		ERRLOG((LOG_ERR,"smfi_register failed"));
		exit(1);
	}
	makepidfile(gcfg.pidfile);
	init_plugins();
	init_dyn_modules();
	rc=smfi_main();
	term_plugins();
	DPRINT(("process %d terminating with rc=%d\n",getpid(),rc));
	ERRLOG((LOG_INFO,"process terminating rc=%d",rc));
#ifdef HAVE_SYSLOG
	closelog();
#endif
	return rc;
}
